/*
簡易HTMLパーサ
XMLとして破綻しているHTMLをXMLパーサを通せる程度に構造化する。
入力はUTF-8を想定
OCaml 4.02
2015.02.11 - 2015.04.04 Takeshi Yoneki
2020.07.05 ロイター向けに改変
2021.02.27 東京新聞のバグ対処
2021.08.05 テレビ王国の間違った閉じタグ
2021.08.10 DOCTYPEを出力できるようにするため、Document型を用意
2021.08.12 整形出力機能
2021.08.31 終了タグ判定関連の改変
2021.09.05 シングルクオート対応など
2022.01.23 Rustへ移植
ostra@asahi.email.ne.jp
*/
use crate::char_entity;
use std::collections::HashSet;

#[derive(Debug)]
pub enum Node {
    Element(String, Vec<(String, String)>, Vec<Node>),
    PCData(String),
    CData(String),
    Comment(String),
    DocType(String),
    Decl(String),
    Document(Vec<Node>),
    Empty,
}

#[derive(Debug)]
enum PartKind {
    StartTag, EndTag, EmptyTag, Text, Comment, Doctype, Decl, CData,
}

fn is_space(ch: char) -> bool {
    ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r'
}

struct ParsePhase1 {
    part: Vec<char>,
    pool: Vec<(PartKind, String)>,
    start_less_than: bool
}

impl ParsePhase1 {
    fn new() -> ParsePhase1 {
        ParsePhase1 { part: Vec::new(), pool: Vec::new(), start_less_than: false }
    }

    fn add_part_to_pool(&mut self, row: String) {
        let s = row.to_lowercase();
        let k = 
            if self.part[0] == '<' && self.part[self.part.len() - 1] == '>' {
                // ここに来る場合、文字列長は必ず2以上なのでpart[1]は存在する。 
                if self.part[1] == '/' {
                    PartKind::EndTag
                }
                else if self.part[1] == '!' {
                    if s.starts_with("<!doctype") {
                        PartKind::Doctype
                    }
                    else if s.starts_with("<![cdata[") && s.ends_with("]]>") {
                        PartKind::CData
                    }
                    else if s.starts_with("<!--") && s.ends_with("-->") {
                        PartKind::Comment
                    }
                    else {
                        PartKind::StartTag
                    }
                }
                // ここに来る場合、キャラクタ数は必ず2以上なのでpart[part.len() - 2]は存在する。 
                else if self.part[self.part.len() - 2] == '/' {
                    PartKind::EmptyTag
                }
                else if self.part[1] == '?' {
                    if self.part[self.part.len() - 2] == '?' {
                        PartKind::Decl
                    }
                    else {
                        PartKind::StartTag
                    }
                }
                else {
                    PartKind::StartTag
                }
            }
            else {
                PartKind::Text
            };
        self.pool.push((k, row));
        self.part.clear();
        self.start_less_than = false;
    }

    fn parse_phase1_proc(&mut self, data: &str) {
        let mut in_tag = false;
        let mut in_quote = '\x00';
        let mut in_attr_value = false;
        let mut in_comment = false;
        let mut in_cdata = false;
        let mut in_script = false;
        //let mut start_less_than = false;

        for ch in data.chars() {
            if ch == '\r' {
            }
            else if in_cdata {
                self.part.push(ch);
                if ch == '>' {
                    let p = self.part.iter().collect::<String>();
                    if p.ends_with("]]>") {
                        self.add_part_to_pool(p);
                        in_cdata = false;
                        in_tag = false;
                        in_attr_value = false;
                    }
                }
            }
            else if in_comment {
                self.part.push(ch);
                if ch == '>' {
                    let p = self.part.iter().collect::<String>();
                    if p.ends_with("-->") {
                        self.add_part_to_pool(p);
                        in_comment = false;
                        in_tag = false;
                        in_attr_value = false;
                    }
                }
            }
            else if in_script {
                if ch == '<' {
                    if !self.part.is_empty() {
                        self.pool.push((PartKind::Text, self.part.iter().collect()));
                        self.part.clear();
                    }
                    self.start_less_than = true;
                }
                self.part.push(ch);
                if self.start_less_than && (ch == '>' || ch == '-' || ch == '[') {
                    let p = self.part.iter().collect::<String>();
                    let q = p.to_lowercase();
                    if ch == '>' && q == "</script>" {
                        self.add_part_to_pool(p);
                        in_script = false;
                        in_tag = false;
                        in_attr_value = false;
                    }
                    else if ch == '-' && q == "<!--" {
                        in_comment = true;
                    }
                    else if ch == '[' && q == "<![cdata[" {
                        in_cdata = true;
                    }
                }
            }
            else if in_tag {
                if in_quote != '\x00' {
                    if ch == in_quote {
                        in_quote = '\x00';
                        in_attr_value = false;
                    }
                    self.part.push(ch);
                }
                else if ch == '>' {
                    self.part.push(ch);
                    let p = self.part.iter().collect::<String>();
                    let q = p.to_lowercase();
                    // part[0] = '<'ならばキャラクタ数は必ず2以上なのでpart[part.len() - 2]は存在する。
                    if self.part[0] == '<' && self.part[self.part.len() - 2] != '/' {
                        if q.starts_with("<script ") || q == "<script>" {
                            in_script = true;
                        }
                    }
                    self.add_part_to_pool(p);
                    in_tag = false;
                    in_attr_value = false;
                }
                else if in_attr_value && is_space(ch) {
                    // 「=」後の空白を詰める。属性値のとき空白前の文字が「=」でなければ属性値終了。
                    // タグの中なら必ずpart[0]は存在する。
                    if self.part[self.part.len() - 1] == '=' {
                    }
                    else {
                        in_attr_value = false;
                        self.part.push(ch);
                    }
                }
                else if ch == '=' {
                    in_attr_value = true;
                    self.part.push(ch);
                }
                else if ch == '\"' || ch == '\'' {
                    // 東京新聞で間違ったダブルクオートが混入している。
                    // タグの中でのクオート終了直後のクオート開始は無視する。
                    // 2021.02.27
                    if !in_attr_value {
                    }
                    else {
                        in_quote = ch;
                        self.part.push(ch);
                    }
                }
                else {
                    self.part.push(ch);
                }
                //
                if ch == '-' {
                    let p = self.part.iter().collect::<String>();
                    if p == "<!--" {
                        in_comment = true;
                    }
                }
                else if ch == '[' {
                    let p = self.part.iter().collect::<String>().to_lowercase();
                    if p == "<![cdata[" {
                        in_cdata = true;
                    }
                }
            }
            else {
                if ch == '<' {
                    if !self.part.is_empty() {
                        let p = self.part.iter().collect::<String>();
                        self.add_part_to_pool(p);
                    }
                    in_tag = true;
                }
                self.part.push(ch);
            }
        }
        if !self.part.is_empty() {
            let p = self.part.iter().collect::<String>();
            self.add_part_to_pool(p);
        }
    }
}

struct ParseStartTag {
    part: Vec<char>,
    pool: Vec<(String, String)>,
    name: String,
    in_value: bool,
}

impl ParseStartTag {
    fn new() -> ParseStartTag {
        ParseStartTag { 
            part: Vec::new(), 
            pool: Vec::new(), 
            name: String::new(), 
            in_value: false,
        }       
    }

    fn enter(&mut self, force: bool) {
        if !self.part.is_empty() || force {
            let r = self.part.iter().collect::<String>();
            //println!("{}", r);
            self.part.clear();
            if self.in_value {
                self.pool.push((self.name.clone(), r));
                self.name.clear();
            }
            else {
                if !self.name.is_empty() {
                    self.pool.push((self.name.clone(), self.name.clone()));
                }
                self.name = r;
            }
        }
    }

    fn parse_start_tag_proc(&mut self, k: &PartKind, row: String) -> Node {
        let mut in_quote = '\x00';
        let sub = 
            match k {
                PartKind::EmptyTag => String::from(&row[1..(row.len() - 2)]),
                _ => String::from(&row[1..(row.len() - 1)]),
            };
        for ch in sub.chars() {
            if in_quote != '\x00' {
                if ch == in_quote {
                    self.enter(true);
                    in_quote = '\x00';
                    self.in_value = false;
                }
                else {
                    self.part.push(ch);
                }
            }
            else {
                if ch == '\"' || ch == '\'' {
                    self.enter(false);
                    in_quote = ch;
                }
                else if ch == '=' {
                    self.enter(false);
                    self.in_value = true;
                }
                else if is_space(ch) {
                    self.enter(false);
                    self.in_value = false;
                }
                else {
                    self.part.push(ch);
                }
            }
        }
        self.enter(false);
        if !self.name.is_empty() {
            self.pool.push((self.name.clone(), self.name.clone()));
        }
        //println!("{:?}", st.pool);
        if !self.pool.is_empty() {
            let (tag, _) = self.pool.remove(0);
            let attr: Vec<(String, String)> = self.pool.drain(..).map(
                |(n, v)| {
                    (n, Parser::fix_char_entity(v))
                }
            ).collect();
            //println!("{:?}", tag);
            Node::Element(tag, attr, Vec::new())
        }
        else {
            Node::Empty
        }
    }
}

struct ParsePhase3 {
    end_tags: HashSet<String>,
    stack: Vec<String>,
}
impl ParsePhase3 {
    fn new(data: &Vec<(PartKind, Node)>) -> ParsePhase3 {
        let et: HashSet<String> = data.iter().filter_map(
            |(k, p)| {
                if matches!(k, PartKind::EndTag) {
                    match p {
                        Node::Element(tag, _, _) => Some(tag.to_lowercase()),
                        _ => panic!("ERROR_BUG")
                    }
                }
                else {
                    None
                }
            }
        ).collect();
        ParsePhase3 {
            end_tags: et, 
            stack: Vec::new(), 
        }
    }

    fn has_end1(&self, tn: &String) -> bool {
        self.end_tags.contains(tn)
    }

    fn mk_inner_data<'a> (tn: &String, mut data: Vec<(PartKind, Node)>) -> (Vec<(PartKind, Node)>, Vec<(PartKind, Node)>, bool) {
        let mut bracket = 1;
        let mut inner: Vec<(PartKind, Node)> = Vec::new();
        while bracket > 0 && !data.is_empty() {
            let d = data.remove(0);
            let (k, p) = &d;
            if let Node::Element(tag, _, _) = p {
                if &tag.to_lowercase() == tn {
                    match k {
                        PartKind::StartTag => bracket += 1,
                        PartKind::EndTag => bracket -= 1,
                        _ => {},
                    }
                }
            }
            inner.push(d);
        }
        if bracket == 0 {
            (data, inner, true)
        }
        else {
            // タグが閉じてない場合は失敗扱いにする。
            // mk_inner_dataはdxを消費する。失敗時はdxを復元する必要がある。
            inner.append(&mut data);
            (inner, Vec::new(), false)
        }
    }

    fn mk_tree(&mut self, data: Vec<(PartKind, Node)>) -> (Vec<(PartKind, Node)>, Vec<Node>, String) {
        let mut dx = data;
        let mut break_flag = false;
        let mut sibling: Vec<Node> = Vec::new();
        let mut q = String::new();
        while !dx.is_empty() && break_flag == false {
            let (k, p) = dx.remove(0);
            match k {
                PartKind::StartTag => {
                    if let Node::Element(tag, attr, _) = &p {
                        let tn = tag.to_lowercase();
                        if self.has_end1(&tn) && dx.len() > 0 {
                            let (mut rest1, inner, has_end2) = ParsePhase3::mk_inner_data(&tn, dx);
                            if has_end2 {
                                self.stack.push(tn.clone());
                                let (mut rest2, child, cue) = self.mk_tree(inner);
                                rest2.append(&mut rest1);
                                dx = rest2;
                                self.stack.pop();
                                let e = Node::Element(String::from(tag), attr.clone(), child);
                                sibling.push(e);
                                if !self.stack.is_empty() && cue != tn && self.stack.contains(&cue) {
                                    q = cue;
                                    break_flag = true;
                                }
                            }
                            else {
                                dx = rest1;
                                sibling.push(p);
                            }
                        }
                        else {
                            sibling.push(p);
                        }
                    }
                    else {
                        //panic!("ERROR_BUG 2");
                    }
                },
                PartKind::EndTag => {
                    if let Node::Element(tag, _, _) = &p {
                        let tn = tag.to_lowercase();
                        let current_tag = 
                            if let Some(t) = self.stack.last() {
                                t
                            }
                            else {
                                ""
                            };
                        // テレビ王国の間違った閉じタグ 2021.08.05
                        if tn != current_tag {
                            let e = Node::Comment(String::from("/") + &tag);
                            sibling.push(e);
                        }
                        else {
                            q = tn;
                            break_flag = true;
                        }
                    }
                    else {
                        panic!("ERROR_BUG 3")
                    }
                },
                _ => {
                    sibling.push(p);
                }
            }
        }
        (dx, sibling, q)
    }
}

#[derive(Debug)]
pub struct Parser {
}

impl Parser {
    fn new() -> Parser {
        Parser {
        }
    }

    fn parse_phase1(&self, data: &str) -> Vec<(PartKind, String)> {
        let mut p1 = ParsePhase1::new();
        p1.parse_phase1_proc(data);
        p1.pool
    }

    fn fix_char_entity(tx: String) -> String {
        let mut t = tx;
        let mut ac: Vec<String> = Vec::new();
        loop {
            let (n, p) = char_entity::find_char_entity(&t);
            if n < 0 {
                ac.push(t);
                break;
            }
            else {
                let n = n as usize;
                let tc: Vec<char> = t.chars().collect();
                let pc: Vec<char> = p.chars().collect();
                let pl = pc.len();
                ac.push(String::from(&tc[0..n].iter().collect::<String>()));
                ac.push(String::from(char_entity::decode_char_entity(&p)));
                t = String::from(&tc[(n + pl)..].iter().collect::<String>());
            }
        }
        ac.concat()
    }

    fn parse_start_tag(&self, k: &PartKind, row: String) -> Node {
        let mut st = ParseStartTag::new();
        //let mut st = ParseStartTag::new(&self);
        st.parse_start_tag_proc(k, row)
    }

    fn parse_end_tag(&self, row: String) -> Node {
        let t: Vec<char> = row.chars().collect();
        let tag = t[2..(t.len() - 1)].iter().collect::<String>();
        Node::Element(tag, Vec::new(), Vec::new())
    }

    fn parse_cdata(&self, row: String) -> Node {
        let t: Vec<char> = row.chars().collect();
        Node::CData(t[9..(t.len() - 3)].iter().collect::<String>())
    }

    fn fix_comment(&self, row: String) -> String {
        // コメント内に"--"を記載してしまうページがある
        let t: Vec<char> = row.chars().collect();
        let s = t[4..(t.len() - 3)].iter().collect::<String>();
        s.replace("--", "&#x2d;&#x2d;")
    }

    fn parse_phase2(&self, mut data: Vec<(PartKind, String)>) -> Vec<(PartKind, Node)> {
        data.drain(..).map(
            |(k, row)| {
                let node = 
                    match &k {
                        PartKind::StartTag | PartKind::EmptyTag => {
                            self.parse_start_tag(&k, row)
                        },
                        PartKind::EndTag => {
                            self.parse_end_tag(row)
                        },
                        PartKind::Text => {
                            Node::PCData(Parser::fix_char_entity(row))
                        },
                        PartKind::CData => {
                            self.parse_cdata(row)
                        },
                        PartKind::Comment => {
                            Node::Comment(self.fix_comment(row))
                        },
                        PartKind::Doctype => {
                            Node::DocType(row)
                        },
                        PartKind::Decl => {
                            Node::Decl(row)
                        },
                        // _ => {
                        //     Node::PCData(row)
                        // }
                    };
                (k, node)
            }
        ).collect()
    }

    fn parse_phase3(&self, data: Vec<(PartKind, Node)>) -> Vec<Node> {
        let mut p3 = ParsePhase3::new(&data);
        let (_, s, _) = p3.mk_tree(data);
        s
    }

    fn parse(&self, data: &str) -> Node {
        let r1 = self.parse_phase1(data);
        // println!("{:?}", r1);
        let r2 = self.parse_phase2(r1);
        //println!("{:?}", r2);
        let r3 = self.parse_phase3(r2);
        //println!("{:?}", p3);
        Node::Document(r3)
    }
}

pub fn parse(data: &str) -> Node {
    Parser::new().parse(data)
}

impl Node {
    pub fn clone(&self) -> Node {
        match self {
            Node::Element(t, at, ch) => Node::Element(t.clone(), at.clone(), ch.iter().map(|c| c.clone()).collect()),
            Node::Document(ch) => Node::Document(ch.iter().map(|c| c.clone()).collect()),
            Node::CData(t) => Node::CData(t.clone()),
            Node::PCData(t) => Node::PCData(t.clone()),
            Node::Comment(t) => Node::Comment(t.clone()),
            Node::DocType(t) => Node::DocType(t.clone()),
            Node::Decl(t) => Node::Decl(t.clone()),
            Node::Empty => Node::Empty,
        }
    }

    pub fn to_xml_string(&self) -> String {
        // 実体参照にすべき文字は実体参照にする
        fn fix_text(t: &str) -> String {
            let mut u = t.replace("&", "&amp;");
            u = u.replace(">", "&gt;");
            u.replace("<", "&lt;")
        }

        fn fix_attr_value(t: &str) -> String {
            let mut u = fix_text(t);
            u = u.replace("\"", "&quot;");
            u.replace("'", "&quos;")
        }

        fn concat_vs(vs: Vec<String>, sep: &str) -> String {
            let mut result = String::new();
            for s in vs {
                if !result.is_empty() {
                    result += sep;
                }
                result += &s;
            }
            result
        }

        fn on_node(e: &Node, mut ac: Vec<String>) -> Vec<String> {
            match e {
                Node::Element(tag, attr, child) => {
                    ac.push(String::from("<") + &tag);
                    let axs: Vec<String> = attr.iter().map(
                        |(n, v)|{
                            String::from(n) + "=\"" + &fix_attr_value(&v) + "\""
                        }).collect();
                    let ax = concat_vs(axs, " ");
                    if !ax.is_empty() {
                        ac.push(String::from(" ") + &ax);
                    }
                    if child.is_empty() {
                        ac.push(String::from(" />"));
                    }
                    else {
                        ac.push(String::from(">"));
                        ac = child.iter().fold(
                            ac,
                            |ac, e| {
                                on_node(&e, ac)
                            }
                        );
                        ac.push(String::from("</") + &tag + ">");
                    }
                    ac
                },
                Node::Comment(t) => {
                    ac.push(String::from("<!--") + &t + "-->");
                    ac
                },
                Node::PCData(t) | Node::CData(t) => {
                    ac.push(fix_text(&t));
                    ac
                },
                Node::DocType(t) | Node::Decl(t) => {
                    ac.push(String::from(t));
                    ac
                },
                Node::Document(ts) => {
                    ts.iter().fold(
                        ac,
                        |ac, e| {
                            on_node(&e, ac)
                        }
                    )
                }
                _ => {
                    ac
                }
            }
        }
        let mut texts = on_node(&self, Vec::new());
        match self {
             Node::Document(_) => {
             }
             _ => {
                texts.push(String::from("\n"));
             }
        }
        texts.concat()
        //on_node(&self, Vec::new()).concat()
    }

    pub fn is_element(&self) -> bool {
        matches!(self, Node::Element(_, _, _))
    }

    pub fn is_text(&self) -> bool {
        matches!(self, Node::PCData(_) | Node::CData(_))
    }

    pub fn set_attribute(self, name: &str, value: &str) -> Node {
        match self {
            Node::Element(tag, mut attr, child) => {
                let mut ax: Vec<(String, String)> = attr.drain(..).filter(|(n, _)| { n != name }).collect();
                ax.push((String::from(name), String::from(value)));
                Node::Element(tag, ax, child)
            },
            _ => {
                self
            }
        }
    }

    pub fn remove_attribute(self, name: &str) -> Node {
        match self {
            Node::Element(tag, mut attr, child) => {
                let ax: Vec<(String, String)> = attr.drain(..).filter(
                    |(n, _)| { 
                        n != name 
                    }
                ).collect();
                Node::Element(tag, ax, child)
            },
            _ => {
                self
            }
        }
    }

    // pub fn remove_attribute_ignore_ns(self, name: &str) -> Node {
    //     match self {
    //         Node::Element(tag, mut attr, child) => {
    //             let ax: Vec<(String, String)> = attr.drain(..).filter(
    //                 |(n, _)| {
    //                     !(n.split(":").last().unwrap() == name)
    //                 }
    //             ).collect();
    //             Node::Element(tag, ax, child)
    //         },
    //         _ => {
    //             self
    //         }
    //     }
    // }

    pub fn remove_attribute_all(self) -> Node {
        match self {
            Node::Element(tag, attr, child) => {
                Node::Element(tag, Vec::new(), child)
            },
            _ => {
                self
            }
        }
    }

    pub fn get_attribute(&self, name: &str) -> String {
        match self {
            Node::Element(_, attr, _) => {
                let ax: Vec<&(String, String)> = attr.iter().filter(|(n, _)| { n == name }).collect();
                if !ax.is_empty() {
                    let (_, v) = &ax[0];
                    return String::from(v);
                }
            },
            _ => {
            }
        }
        return String::new();
    }

    // pub fn get_attributes(&self) -> Vec<(String, String)> {
    //     match self {
    //         Node::Element(_, attr, _) => {
    //             let attr2 = attr.iter().map(
    //                 |(name, value)| {
    //                     (String::from(name), String::from(value))
    //                 }
    //             ).collect();
    //             return attr2;
    //         },
    //         _ => {
    //         }
    //     }
    //     return Vec::new();
    // }

    pub fn get_document_element(&self) -> &Node {
        match self {
            Node::Document(child) => {
                for n in child {
                    if n.is_element() {
                        return n;
                    }
                }
                return &Node::Empty;
            },
            Node::Element(_, _, _) => {
                self
            },
            _ => {
                &Node::Empty
            }
        }
    }

    fn get_data_text(&self) -> String {
        match self {
            Node::PCData(t) | Node::CData(t) => String::from(t),
            _ => String::new()
        }
    }

    fn get_element_text(&self) -> String {
        match self {
            Node::Element(_, _, child) => {
                let ts: Vec<String> = child.iter().map(
                    |n| {
                        n.get_data_text()
                    }
                ).collect();
                return ts.concat();
            },
            _ => {
            }
        }
        return String::new();
    }

    // pub fn set_elmenet_text(self, text: &str) -> Node {
    //     match self {
    //         Node::Element(tag, attr, child) => {
    //             Node::Element(tag, attr, vec![Node::PCData(String::from(text))])
    //         },
    //         _ => {
    //             self
    //         }
    //     }
    // }

    pub fn get_text_content(&self) -> String {
        match self {
            Node::Element(_, _, _) => self.get_element_text(),
            _ => self.get_data_text(),
        }
    }

    pub fn get_text_content_recur(&self) -> String {
        match self {
            Node::Element(_, _, child) => {
                child.iter().fold(String::new(),
                    |acc, n| {
                        acc + &n.get_text_content_recur()
                    }
                )
            },
            _ => {
                self.get_data_text()
            }
        }
    }

    pub fn get_elements_by_tag_name(&self, name: &str) -> Vec<&Node> {
        match self {
            Node::Element(tag, _, child) => {
                let mut ns: Vec<&Node> = child.iter().flat_map(
                    |n|{
                        n.get_elements_by_tag_name(name)
                    }
                ).collect();
                if tag == name || tag.split(":").last().unwrap() == name {
                    //println!("A {} {:?}", name, self);
                    ns.insert(0, self);
                }
                ns
            },
            Node::Document(child) => {
                let ns: Vec<&Node> = child.iter().flat_map(
                    |n|{
                        n.get_elements_by_tag_name(name)
                    }
                ).collect();
                ns
            },
            _ => {
                Vec::new()
            }
        }
    }

    pub fn get_children(&self) -> Vec<&Node> {
        match self {
            Node::Element(_, _, child) => child.iter().collect(),
            _ => Vec::new()
        }
    }

    pub fn get_child(&self) -> Vec<&Node> {
        self.get_children()
    }

    pub fn get_tag_name(&self) -> String {
        match self {
            Node::Element(tag, _, _) => String::from(tag),
            _ => String::new()
        }
    }

    pub fn append_child(self, n: Node) -> Node {
        match self {
            Node::Element(tag, attr, mut child) => {
                child.push(n);
                Node::Element(tag, attr, child)
            },
            _ => {
                self
            }
        }
    }

    fn rule_transformer_sub<F>(self, trans: &mut F, path: &str) -> Node
        where F: FnMut(Node, &str) -> Node {
        let p = 
            if self.is_element() {
                String::from(path) + "/" + &self.get_tag_name()
            }
            else {
                String::from(path)
            };
        let r = trans(self, &p);
        match r {
            Node::Element(tag, attr, mut child) => {
                let ns: Vec<Node> = child.drain(..).map(
                    |n| {
                        n.rule_transformer_sub(trans, &p)
                    }
                ).collect();
                Node::Element(tag, attr, ns)
            },
            Node::Document(mut child) => {
                let ns: Vec<Node> = child.drain(..).map(
                    |n| {
                        n.rule_transformer_sub(trans, &p)
                    }
                ).collect();
                Node::Document(ns)
            },
            _ => {
                r
            }
        }
    }

    pub fn rule_transformer<F>(self, trans: &mut F) -> Node
        where F: FnMut(Node, &str) -> Node {
        self.rule_transformer_sub(trans, "")
    }

    pub fn pretty(self, tab: &str) -> Node {
        fn is_space_string(s: &str) -> bool {
            let t:Vec<char> = s.chars().filter(
                |ch| {
                    !is_space(*ch)
                }
            ).collect();
            t.is_empty()
        }

        fn has_text_pcdata(child: &Vec<Node>) -> bool {
            let child2: Vec<&Node> = child.iter().filter(
                |n| {
                    match n {
                        Node::PCData(t) => !is_space_string(t),
                        Node::CData(t) => !is_space_string(t),
                        _ => false,
                    }
                }
            ).collect();
            !child2.is_empty()
        }

        fn delete_space_pcdata(mut child: Vec<Node>) -> Vec<Node> {
            if child.is_empty() || has_text_pcdata(&child) {
                child
            }
            else {
                child.drain(..).filter(
                    |n| {
                        match n {
                            Node::PCData(t) => !is_space_string(t),
                            Node::CData(t) => !is_space_string(t),
                            _=> true,
                        }
                    }
                ).collect()
            }
        }

        fn delete_space(doc: Node) -> Node {
            let doc2 = doc.rule_transformer(
                &mut |node, _| {
                    match node {
                        Node::Element(tag, attr, child) =>
                            Node::Element(tag, attr, delete_space_pcdata(child)),
                        Node::Document(child) =>
                            Node::Document(delete_space_pcdata(child)),
                        _ => node,
                    }
                }
            );
            doc2
        }

        fn calc_indent_count(path: &str) -> usize {
            let t: Vec<char> = path.chars().filter(
                |ch| {
                    ch == &'/'
                }
            ).collect();
            t.len()
        }

        fn mk_indent_pcdata(tab: &str, count: usize) -> Node {
            if count < 1 {
                Node::PCData(String::new())
            }
            else {
                let t: Vec<String> = (0..count).fold(
                    Vec::new(),
                    |mut acc, _| {
                        acc.push(String::from(tab));
                        acc
                    }
                );
                Node::PCData(t.concat())
            }
        }

        fn indent_space(doc: Node, tab: &str) -> Node {
            fn mk_endline() -> Node {
                Node::PCData(String::from("\n"))
            }
            let doc2 = doc.rule_transformer(
                &mut |node, path| {
                    match node {
                        Node::Element(tag, attr, mut child) => {
                            if child.is_empty() || has_text_pcdata(&child) {
                                Node::Element(tag, attr, child)
                            }
                            else {
                                let indent_count = calc_indent_count(path);
                                let mut child2 = child.drain(..).fold(
                                    vec![mk_endline()],
                                    |mut acc, n| {
                                        acc.push(mk_indent_pcdata(tab, indent_count));
                                        acc.push(n);
                                        acc.push(mk_endline());
                                        acc
                                    }
                                );
                                child2.push(mk_indent_pcdata(tab, indent_count - 1));
                                Node::Element(tag, attr, child2)
                            }
                        },
                        Node::Document(mut child) => {
                            let child2 = child.drain(..).fold(
                                Vec::new(),
                                |mut acc, n| {
                                    acc.push(n);
                                    acc.push(mk_endline());
                                    acc
                                }
                            );
                            Node::Document(child2)
                        },
                        _ => {
                            node
                        }
                    }
                }
            );
            doc2
        }
        //
        let doc2 = delete_space(self);
        let doc3 = indent_space(doc2, tab);
        doc3
    }
}
